%{
/*
 * Booking dates form generator - template parser
 * Copyright (C) 1999 Open World Ltd
 * Written by Shevek <shevek@anarres.org>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, 
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the 
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License 
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <time.h>
#include <stdio.h>
#include <unistd.h>
#include <string.h>
#include <errno.h>
#include "library.h"
#include "rope.h"
#include "cgi.h"
#include "smalloc.h"
#include "log.h"

/* ts_count will loop up to these values */
#define N_YEAR	(2)
#define N_MONTH	(12)
#define N_DAY	(31)
#define N_ALL	(1)		/* Unimplementable? */

typedef
struct _tpt_state {
	struct tm	*ts_tm;

	int			 ts_count;
	int			 ts_max;

	int			 ts_select;
} tpt_state;

typedef
struct _profile {
	char	*dayfmt;	/* printf format */
	int		 monstrp;	/* Is month a string or int */
	char	*monfmt;	/* printf format */
	char	*yearfmt;	/* printf format */
} profile;

/* If strlen(format) == 0, don't print */

#define ENG_TRAVELWEB	(0)
profile engines[] =
{
	{ "%.2d", 0, "%.2d", "" }
};

profile	 engine;

static void template_buffer_start(struct tm *tm, int maxloop, int sel);
static void template_buffer_parse(void);

/* Old protocol */
static void tk_day_nlz(void);
static void tk_day_lz(void);
static void tk_day_ord(void);
static void tk_mon_nlz(void);
static void tk_mon_lz(void);
static void tk_mon_card(void);
static void tk_mon_short(void);
static void tk_mon_long(void);
static void tk_yr_short(void);
static void tk_yr_long(void);
static void tk_selected(void);

/* New protocol */
static void prf_year(struct tm *tm);
static void prf_monthyear(struct tm *tm);
static void prf_month(struct tm *tm);
static void prf_day(struct tm *tm);

struct tm	*today;
struct tm	*tomorrow;

rope_t		*buffer = NULL;

tpt_state	 state;

%}

%option noyywrap
%option nounput

%x	BUFFER
%x	TOKEN

%x	CONFIG
%x	PROFILE

	/* Global blocks */
	/* Block yearMonthDaysPresent start */
tBL_YMD_ST		\{yearMonthDaysPresent\}
tBL_YMD_END		\{\/yearMonthDaysPresent\}

tBL_YMAD_ST		\{yearMonthsAndDaysPresent\}
tBL_YMAD_END	\{\/yearMonthsAndDaysPresent\}

tBL_YAMAD_ST	\{yearsAndMonthsAndDaysPresent\}
tBL_YAMAD_END	\{\/yearsAndMonthsAndDaysPresent\}

	/* Data format blocks */
	/* YMD blocks */
tFB_ARR_YMD		\{arrival_yearMonthDays\}
tFB_DEP_YMD		\{departure_yearMonthDays\}
	/* YMAD blocks */
tFB_ARR_YM		\{arrival_yearMonths\}
tFB_DEP_YM		\{departure_yearMonths\}
	/* YAMAD blocks */
tFB_ARR_YEAR	\{arrival_years\}
tFB_ARR_MON		\{arrival_months\}
tFB_ARR_DAY		\{arrival_days\}
tFB_DEP_YEAR	\{departure_years\}
tFB_DEP_MON		\{departure_months\}
tFB_DEP_DAY		\{departure_days\}

	/* Data format blocks */
	/* YMD blocks */
eFB_ARR_YMD		\{\/arrival_yearMonthDays\}
eFB_DEP_YMD		\{\/departure_yearMonthDays\}
	/* YMAD blocks */
eFB_ARR_YM		\{\/arrival_yearMonths\}
eFB_DEP_YM		\{\/departure_yearMonths\}
	/* YAMAD blocks */
eFB_ARR_YEAR	\{\/arrival_years\}
eFB_ARR_MON		\{\/arrival_months\}
eFB_ARR_DAY		\{\/arrival_days\}
eFB_DEP_YEAR	\{\/departure_years\}
eFB_DEP_MON		\{\/departure_months\}
eFB_DEP_DAY		\{\/departure_days\}

	/* Printing tokens */
	/* Token (printing) day no leading zero */
tTK_DAY_NLZ		\{day_noLeadingZero\}
tTK_DAY_LZ		\{day_leadingZero\}
tTK_DAY_ORD		\{day_ordinal\}

	/* Ordinal, cardinal */
tTK_MON_NLZ		\{month_noLeadingZero\}
tTK_MON_LZ		\{month_leadingZero\}
tTK_MON_CARD	\{month_zeroBasedNoLeadingZero\}
tTK_MON_SHORT	\{month_shortName\}
tTK_MON_LONG	\{month_longName\}

tTK_YR_SHORT	\{year_short\}
tTK_YR_LONG		\{year_long\}

tTK_SELECTED	\{selected\}

tPRF_CONFIG		\{config\}
ePRF_CONFIG		\{\/config\}

tPRF_ARR_D		\{arrival_day\}
tPRF_DEP_D		\{departure_day\}
tPRF_ARR_M		\{arrival_month\}
tPRF_DEP_M		\{departure_month\}
tPRF_ARR_Y		\{arrival_year\}
tPRF_DEP_Y		\{departure_year\}
tPRF_ARR_MY		\{arrival_monthyear\}
tPRF_DEP_MY		\{departure_monthyear\}


%%

	/* Ignore these, they're over-syntactic */
{tBL_YMD_ST}	{ }
{tBL_YMD_END}	{ }
{tBL_YMAD_ST}	{ }
{tBL_YMAD_END}	{ }
{tBL_YAMAD_ST}	{ }
{tBL_YAMAD_END}	{ }

	/* These start buffering */
{tFB_ARR_YMD}	{ template_buffer_start(today, N_ALL, 0);		}
{tFB_ARR_YM}	{ template_buffer_start(today, N_MONTH, 0);		}

{tFB_DEP_YMD}	{ template_buffer_start(tomorrow, N_ALL, 0);	}
{tFB_DEP_YM}	{ template_buffer_start(tomorrow, N_MONTH, 0);	}

{tFB_ARR_YEAR}	{ template_buffer_start(today, N_YEAR, 0);		}
{tFB_ARR_MON}	{ template_buffer_start(today, N_MONTH, 0);		}
{tFB_ARR_DAY}	{ template_buffer_start(today, N_DAY, 1);		}

{tFB_DEP_YEAR}	{ template_buffer_start(tomorrow, N_YEAR, 0);	}
{tFB_DEP_MON}	{ template_buffer_start(tomorrow, N_MONTH, 0);	}
{tFB_DEP_DAY}	{ template_buffer_start(tomorrow, N_DAY, 1);	}

<INITIAL>[^\{\r\n]+		|
<INITIAL>\{				{ ECHO; }

	/* These start parsing */
<BUFFER>{eFB_ARR_YMD}	|
<BUFFER>{eFB_ARR_YM}	|

<BUFFER>{eFB_ARR_YEAR}	|
<BUFFER>{eFB_ARR_MON}	|
<BUFFER>{eFB_ARR_DAY}	|

<BUFFER>{eFB_DEP_YMD}	|
<BUFFER>{eFB_DEP_YM}	|

<BUFFER>{eFB_DEP_YEAR}	|
<BUFFER>{eFB_DEP_MON}	|
<BUFFER>{eFB_DEP_DAY}	{ template_buffer_parse();	}

	/* There may come a time where this misses newlines so put in
	 * explicit newline buffering foo bars */
<BUFFER>[^\{\r\n]+		|
<BUFFER>\{				|
<BUFFER>\n				|
<BUFFER>\r				{ rope_add(buffer, yytext);		}

	/* Token printing */
<TOKEN>{tTK_DAY_NLZ}	{ tk_day_nlz();					}
<TOKEN>{tTK_DAY_LZ}		{ tk_day_lz();					}
<TOKEN>{tTK_DAY_ORD}	{ tk_day_ord();					}

	/* Ordinal, cardinal */
<TOKEN>{tTK_MON_NLZ}	{ tk_mon_nlz();					}
<TOKEN>{tTK_MON_LZ}		{ tk_mon_lz();					}
<TOKEN>{tTK_MON_CARD}	{ tk_mon_card();				}
<TOKEN>{tTK_MON_SHORT}	{ tk_mon_short();				}
<TOKEN>{tTK_MON_LONG}	{ tk_mon_long();				}

<TOKEN>{tTK_YR_SHORT}	{ tk_yr_short();				}
<TOKEN>{tTK_YR_LONG}	{ tk_yr_long();					}

<TOKEN>{tTK_SELECTED}	{ tk_selected();				}

<INITIAL>{tPRF_CONFIG}	{ BEGIN(CONFIG);				}
<CONFIG>{ePRF_CONFIG}	{ BEGIN(PROFILE);				}
<CONFIG>[^\{\r\n]+		{ }
<CONFIG>\{				{ }
	/* XXX We need a profile selection rule in here, and we need to
	 * check that we have selected a profile before we go to PROFILE
	 * state. */

<PROFILE>tPRF_ARR_D
<PROFILE>tPRF_DEP_D
<PROFILE>tPRF_ARR_M
<PROFILE>tPRF_DEP_M
<PROFILE>tPRF_ARR_Y
<PROFILE>tPRF_DEP_Y
<PROFILE>tPRF_ARR_MY
<PROFILE>tPRF_DEP_MY

	/* Remember longest match rules. This will print blank in config */
<*>\r\n					{ ECHO;							}
<*>\r					{ printf("\r\n");				}
<*>\n					{ printf("\r\n");				}

%%

const char *months[] = {
	"January",
	"February",
	"March",
	"April",
	"May",
	"June",
	"July",
	"August",
	"September",
	"October",
	"November",
	"December"
};

void
template_times(void)
{
	time_t		 timebuf;
	struct tm	*tm;

	time(&timebuf);
	today = gmtime(&timebuf);

	/* Keep this one */
	tm = (struct tm *)smalloc(sizeof(struct tm));
	memcpy(tm, today, sizeof(struct tm));
	today = tm;

	timebuf += 86400;
	tomorrow = gmtime(&timebuf);
}

void
template_reset(struct tm *tm, int maxloop, int sel)
{
	state.ts_tm = tm;

	state.ts_count = 0;
	state.ts_max = maxloop;

	state.ts_select = sel;
}

void
template_stream(rope_t *pathstr)
{
	char	*path;

	path = rope_chars(pathstr);

	if (access(path, R_OK) != 0) {
		log("cgiconf", PRI_CRIT, "Could not open template '%s': %s",
						path, strerror(errno));
	}
	yyin = fopen(path, "r");
	if (!yyin) {
		log("cgiconf", PRI_CRIT, "Could not open template '%s': %s",
						path, strerror(errno));
	}
}

void
template_parse(rope_t *pathstr)
{
	template_times();
	template_stream(pathstr);

	printf("Content-type: text/html\r\n\r\n");

	yylex();
}

/* === */

static int
template_buffer_done(void)
{
	if (state.ts_count >= state.ts_max)
		return 1;

	return 0;
}

/* This function's third argument is hackish, but it made the code
 * above fit onto a 78 line screen. I should explicitly pass sel
 * as a value, rather than just saying 'is it tm_mday or not?'
 */
static void
template_buffer_start(struct tm *tm, int maxloop, int dayp)
{
	if (buffer) {
		log("cgisys", PRI_ERR, "Nested repeat block start conditions");
		exit(0);
	}

	buffer = rope_new("");

	template_reset(tm, maxloop, (dayp?tm->tm_mday:0));

	BEGIN(BUFFER);
}

static void
template_buffer_parse(void)
{
	YY_BUFFER_STATE	 topbuf;
	YY_BUFFER_STATE	 subbuf;
	/* Store in global so we can pick up on it when we get out */
	topbuf = YY_CURRENT_BUFFER;

	BEGIN(TOKEN);

/* while */
	while (!template_buffer_done()) {
		subbuf = yy_scan_string(rope_chars(buffer));
		yylex();
		yy_delete_buffer(subbuf);

		state.ts_count++;
	}
/* wend */

	rope_delete(buffer);
	buffer = NULL;

	BEGIN(INITIAL);
	yy_switch_to_buffer(topbuf);
}

/* We base on zero internally, but wish to print out values from 1 */
static void
tk_day_nlz(void)
{
	int	 val;

	val = state.ts_count + 1;
	printf("%d", val);
}

static void
tk_day_lz(void)
{
	int	 val;

	val = state.ts_count + 1;
	printf("%.2d", val);
}

static void
tk_day_ord(void)
{
	int	 i;
	int	 val;

	val = state.ts_count + 1;
	printf("%d", val);
	i = val % 10;
	switch(i) {
		case 1: printf("st");
		case 2: printf("nd");
		case 3: printf("rd");
		default: printf("th");
	}
}

static void
tk_mon_nlz(void)
{
	int	 val;
	val = (state.ts_tm->tm_mon + state.ts_count) % 12;
	printf("%d", val + 1);
}

static void
tk_mon_lz(void)
{
	int	 val;
	val = (state.ts_tm->tm_mon + state.ts_count) % 12;
	printf("%.2d", val + 1);
}

static void
tk_mon_card(void)
{
	int	 val;
	val = (state.ts_tm->tm_mon + state.ts_count) % 12;
	printf("%d", val);
}

static void
tk_mon_short(void)
{
	int		 val;
	val = (state.ts_tm->tm_mon + state.ts_count) % 12;
	printf("%.3s", months[val]);
}

static void
tk_mon_long(void)
{
	int		 val;
	val = (state.ts_tm->tm_mon + state.ts_count) % 12;
	printf("%s", months[val]);
}

static void
tk_yr_short(void)
{
	int	 val;
	val = (state.ts_tm->tm_mon + state.ts_count) / 12;	/* 1 or 0 */
	val += state.ts_tm->tm_year + 1900;
	printf("%.2d", val);
}

static void
tk_yr_long(void)
{
	int	 val;
	val = (state.ts_tm->tm_mon + state.ts_count) / 12;	/* 1 or 0 */
	val += state.ts_tm->tm_year + 1900;
	printf("%.4d", val);
}

static void
tk_selected(void)
{
	if (state.ts_select == state.ts_count)
		printf("SELECTED");
}

/* End of old protocol */

static void
prf_day(struct tm *tm)
{
	int	 i;
	for (i = 0; i < N_DAY; i++) {
		printf("<OPTION ");
		if (i == tm->tm_mday)
			printf("SELECTED ");
		if (strlen(engine.dayfmt)) {
			printf("VALUE=");
			printf(engine.dayfmt, i);
		}
		printf("> %.2d\n", i);
	}
}

static void
prf_month(struct tm *tm)
{
	int	 i, j;
	for (i = 0; i < N_MONTH; i++) {
		j = tm->tm_mon + i;
		j %= 12;

		printf("<OPTION ");

		if (i == 0)
			printf("SELECTED ");

		if (strlen(engine.monfmt)) {
			printf("VALUE=");
			if (engine.monstrp)
				printf(engine.monfmt, months[j]);
			else
				printf(engine.monfmt, j);
		}

		printf("> %s\n", months[j]);
	}
}

static void
prf_year(struct tm *tm)
{
	int	 i, j;
	for (i = 0; i < N_YEAR; i++) {
		j = i + tm->tm_year + 1900;

		printf("<OPTION ");
		if (i == 0)
			printf("SELECTED ");
		if (strlen(engine.yearfmt)) {
			printf("VALUE=");
			printf(engine.yearfmt, j);
		}
		printf("> %.4d\n", j);
	}
}

static void
prf_monthyear(struct tm *tm)
{
	int	 i;
	int	 mon;
	int	 year;

	for (i = 0; i < N_MONTH; i++) {
		mon = tm->tm_mon + i;
		year = tm->tm_year + 1900 + (mon / 12);
		mon %= 12;

		printf("<OPTION ");

		if (i == 0)
			printf("SELECTED ");

		if (strlen(engine.monfmt)) {
			printf("VALUE=");
			if (engine.monstrp)
				printf(engine.monfmt, months[mon]);
			else
				printf(engine.monfmt, mon);
		}

		printf("> %s %.4d\n", months[mon], year);
	}
}
