/*
 * C utility library
 * Copyright (C) 1999 Open World Ltd
 * Written by Shevek <shevek@anarres.org>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, 
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the 
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License 
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <stdlib.h>
#include "library.h"
#include "stack.h"
#include "list.h"
#include "smalloc.h"

typedef
struct _stack_t
{
    void *list;
    void *iterator;
} stack_t;

void *
stack_new(stack_delete_t func)
{
    stack_t *stack;
    
    stack = (stack_t *)smalloc(sizeof(stack_t));
    
    stack->list = list_new(func);
    stack->iterator = list_iterator_new(stack->list);
    
    return stack;
}

void
stack_delete(void *stack)
{
    list_iterator_delete(((stack_t *)stack)->iterator);
    list_delete(((stack_t *)stack)->list);
    free(stack);
}

void stack_push(void *stack, void *data)
{
    list_data_insert(((stack_t *)stack)->iterator, NULL, data);
}

void
*stack_pop(void *stack)
{
    void *data;
    void *iterator;
    
    iterator = ((stack_t *)stack)->iterator;
    
    if (list_iterator_prev(iterator))
    {
        data = list_data_read_data(iterator);
        list_data_remove(iterator);
        list_iterator_end(iterator);
    }
    else {
        data = NULL;
    }
    
    return data;
}

bool
stack_isempty(void *stack)
{
    if (list_length(((stack_t *)stack)->list) > 0)
        return FALSE;
    else
        return TRUE;
}
