/*
 * C utility library
 * Copyright (C) 1999 Open World Ltd
 * Written by Shevek <shevek@anarres.org>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, 
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the 
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License 
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include "library.h"
#include "vector.h"
#include "smalloc.h"

void *	/* vector_t */
explode(char *source, char *delim)	/* LPC rexplode() */
{
	static char	*input;
	static int	 len;
	vector_t	*vec;
	char		*basep;
	char		*endp;

	if (!input) {
		len = strlen(source) + 10;
		input = (char *)smalloc(len);
	}
	else if (len <= strlen(source)) {
		len = strlen(source) + 10;
		input = (char *)srealloc(input, len);
	}

	strcpy(input, source);

	vec = vector_new(free, 0);

	basep = input;
	endp = strpbrk(input, delim);
	while (endp) {
		*endp = '\0';
		vector_push_back(vec, sstrdup(basep));
		basep = ++endp;
		endp = strpbrk(basep, delim);
	}
	vector_push_back(vec, sstrdup(basep));

	return vec;
}

char *
implode(vector_t *v, char *delim)
{
	static rope_t	*rope = NULL;
	char			*str;
	int				 len;
	int				 i;

	if (!rope)
		rope = rope_new("");
	else
		rope_set(rope, "");
	
	len = vector_length(v);

	for (i = 0; i < len; i++) {
		if (i) rope_add(rope, delim);
		str = (char *)vector_data(v, i);
		if (!str) str = "(null)";
		rope_add(rope, str);
	}

	str = sstrdup(rope_chars(rope));
	return str;
}

char *
lower_case(char *str)
{
	char	*cp;
	cp = str;
	while (*cp) {
		*cp = tolower(*cp);
		cp++;
	}
	return str;
}
